"""
"""

# Created on 2015.03.18
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

# from threading import Thread
# from multiprocessing import Process
import logging
from .. import EXEC_PROCESS, EXEC_THREAD

from .. import NATIVE_ASYNCIO
from time import sleep

if NATIVE_ASYNCIO:
    import asyncio
else:
    import trollius as asyncio
    from trollius import From, Return


class Instance(object):
    def __init__(self, dsa, name=None, executor=EXEC_THREAD):
        self.dsa = dsa
        self.dsa.instance = self
        self.loop = None
        self.name = self.dsa.name if not name else name
        if executor == EXEC_THREAD:
            from threading import Thread
            self.executor = Thread(target=self.dsa.start)
        elif executor == EXEC_PROCESS:
            from multiprocessing import Process
            self.executor = Process(target=self.dsa.start)
        else:
            raise Exception('unknown executor')

        self.started = False

    def start(self):
        if not self.started:
            logging.info('starting instance %s' % self.name)
            self.executor.start()
            self.started = True

    def stop(self):
        if self.started:
            logging.info('stopping instance %s' % self.name)
            self.dsa.stop()
            logging.debug('stopping loop for instance %s' % self.name)
            self.loop.call_soon_threadsafe(self.loop.stop)
            logging.debug('closing loop for instance %s' % self.name)
            while self.loop.is_running():
                logging.debug('waiting for Instance %s loop to stop' % self.name)
                sleep(0.2)
            self.loop.call_soon_threadsafe(self.loop.close)
            logging.info('Instance %s loop halted and closed' % self.name)
            logging.debug('waiting for instance %s executor to join' % self.name)
            self.executor.join()
            logging.debug('instance %s joined' % self.name)
            self.started = False
            logging.info('stopped instance %s' % self.name)



"""
"""

# Created on 2015.04.15
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import logging
from time import sleep
import sys
from sldap3 import EXEC_THREAD, EXEC_PROCESS

logging.basicConfig(
    filename='c:\\Temp\\sldap3.log',
    level=logging.DEBUG,
    format='[sldap3-service] %(levelname)-7.7s %(message)s'
)

try:
    import win32serviceutil
    import win32service
    import win32event
    import servicemanager
except ImportError:
    logging.error('pywin32 package missing')
    sys.exit(1)

sys.stderr = open('C:\\Temp\\pyasn1.log', 'a')  # patch for pyasn1 without access to stderr

try:
    import pyasn1
except ImportError:
    logging.error('pyasn1 package missing')
    sys.exit(2)

try:
    import ldap3
except ImportError:
    logging.error('ldap3 package missing')
    sys.exit(3)

try:
    from asyncio import BaseEventLoop
except ImportError:
    try:
        import trollius as asyncio
        from trollius import From, Return
    except:
        logging.error('trollius package missing')
        sys.exit(4)

try:
    import sldap3
except ImportError:
    logging.error('sldap3 package missing')
    sys.exit(5)


class Sldap3Service (win32serviceutil.ServiceFramework):
    _svc_name_ = 'sldap3'
    _svc_display_name_ = 'sldap3 - LDAP Server'
    _svc_description_ = 'A strictly RFC 4511 conforming LDAP V3 pure Python server'

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.stop_event = win32event.CreateEvent(None, 0, 0, None)
        self.instances = list()
        self.stop_requested = False

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.stop_event)
        logging.info('stopping sldap3 service...')
        self.stop_requested = True

    def SvcDoRun(self):
        logging.info('running sldap3 service...')
        servicemanager.LogMsg(
            servicemanager.EVENTLOG_INFORMATION_TYPE,
            servicemanager.PYS_SERVICE_STARTED,
            (self._svc_name_, '')
        )
        self.run()
        logging.info('ending sldap3 service...')

    def run(self):
        logging.info('executing sldap3 service...')

        logging.info('instantiating sldap3 daemon')
        self.instances = []
        user_backend = sldap3.JsonUserBackend('/root/sldap3/test/localhost-users.json')
        user_backend.add_user('giovanni', 'admin', 'password')
        user_backend.add_user('beatrice', 'user', 'password')
        user_backend.store()

        dsa1 = sldap3.Instance(
            sldap3.Dsa('DSA1',
                       '0.0.0.0',
                       cert_file='/root/sldap3/test/server-cert.pem',
                       key_file='/root/sldap3/test/server-key.pem',
                       user_backend=user_backend),
            name='MixedInstance',
            executor=EXEC_THREAD)
        dsa2 = sldap3.Instance(
            sldap3.Dsa('DSA2',
                       '0.0.0.0',
                       port=1389,
                       user_backend=user_backend),
            name='UnsecureInstance',
            executor=EXEC_THREAD)

        self.instances.append(dsa1)
        self.instances.append(dsa2)

        for instance in self.instances:  # start each instance in a new thread
            instance.start()

        logging.info('sldap3 daemon instantiation complete')

        while not self.stop_requested:  # wait for stop signal
            sleep(3)

        for instance in self.instances:  # wait for all instances to end
            instance.stop()

        logging.info('sldap3 service stopped')

if __name__ == '__main__':
    win32serviceutil.HandleCommandLine(Sldap3Service)



# THIS FILE IS AUTO-GENERATED. PLEASE DO NOT MODIFY# version file for sldap3
# generated on 2015-04-23 00:20:11.150706
# on system uname_result(system='Windows', node='GCNBHPW8', release='8', version='6.2.9200', machine='AMD64', processor='Intel64 Family 6 Model 58 Stepping 9, GenuineIntel')
# with Python 3.4.3 - ('v3.4.3:9b73f1c3e601', 'Feb 24 2015 22:44:40') - MSC v.1600 64 bit (AMD64)
#
__version__ = '0.0.4a0'
__author__ = 'Giovanni Cannata'
__email__ = 'cannatag@gmail.com'
__url__ = 'https://github.com/cannatag/sldap3'
__description__ = 'A strictly RFC 4511 conforming LDAP V3 pure Python server. Same codebase for Python 2, Python3, PyPy and PyPy 3'
__status__ = 'development - alpha'
__license__ = 'LGPL v3'


"""
"""

# Created on 2015.04.20
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
from sldap3 import EXEC_THREAD

logging.basicConfig(
    filename='/var/log/sldap3.log',
    level=logging.DEBUG,
    format='%(asctime)s - %(name)s - %(levelname)-7.7s - %(message)s'
)

try:
    import resource
except ImportError:
    logging.error('deamons are available on Linux platform only')
    sys.exit(5)

try:
    from pep3143daemon import DaemonContext, PidFile
except ImportError:
    logging.error('pep3143daemon package missing')
    sys.exit(6)

try:
    import pyasn1
except ImportError:
    logging.error('pyasn1 package missing')
    sys.exit(2)

try:
    import ldap3
except ImportError:
    logging.error('ldap3 package missing')
    sys.exit(3)

try:
    from asyncio import BaseEventLoop
except ImportError:
    try:
        import trollius as asyncio
        from trollius import From, Return
    except:
        logging.error('trollius package missing')
        sys.exit(4)

try:
    import sldap3
except ImportError:
    logging.error('sldap3 package missing')
    sys.exit(5)


class Sldap3Daemon(DaemonContext):
    def run(self):
        logging.info('instantiating sldap3 daemon')
        self.instances = []
        user_backend = sldap3.JsonUserBackend('/root/sldap3/test/localhost-users.json')
        user_backend.add_user('giovanni', 'admin', 'password')
        user_backend.add_user('beatrice', 'user', 'password')
        user_backend.store()

        dsa1 = sldap3.Instance(
            sldap3.Dsa('DSA1',
                       '0.0.0.0',
                       cert_file='/root/sldap3/test/server-cert.pem',
                       key_file='/root/sldap3/test/server-key.pem',
                       user_backend=user_backend),
            name='MixedInstance',
            executor=EXEC_THREAD)
        dsa2 = sldap3.Instance(
            sldap3.Dsa('DSA2',
                       '0.0.0.0',
                       port=1389,
                       user_backend=user_backend),
            name='UnsecureInstance',
            executor=EXEC_THREAD)

        self.instances.append(dsa1)
        self.instances.append(dsa2)

        for instance in self.instances:  # start each instance in a new thread
            instance.start()

        logging.info('sldap3 daemon instantiation complete')

    def terminate(self, signal_number, stack_frame):
        logging.info('terminating sldap3 daemon')
        for instance in self.instances:  # wait for all instances to end
            instance.stop()
        logging.info('daemon sldap3 terminated')

if __name__ == '__main__':
    pid = '/tmp/sldap3.pid'
    pidfile = PidFile(pid)
    daemon = Sldap3Daemon(pidfile=pidfile)
    daemon.files_preserve = [logging.getLogger().handlers[0].stream]  # preserve log file
    logging.debug('preserving files %s' % str(daemon.files_preserve))

    logging.info('daemonizing sldap3')
    daemon.open()
    logging.info('sldap3 demonized')
    daemon.run()
    logging.info('sldap3 daemon started')



"""
"""

# Created on 2015.03.13
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import ssl
import logging

from datetime import datetime
from ldap3 import RESULT_PROTOCOL_ERROR
from pyasn1.codec.ber import decoder, encoder
from ..protocol.rfc4511 import build_extended_response, build_ldap_result, build_ldap_message


class Dua(object):
    """
    Directory User Agent - a client actually connected to the DSA with an active transport
    """
    def __init__(self, user, reader, writer, dsa):
        self.user = user
        self.dsa = dsa
        self.connected_time = datetime.now()
        self.reader = reader
        self.writer = writer
        self.tls_started = False
        self.pending = {}

    def send(self, ldap_message):
        encoded_message = encoder.encode(ldap_message)
        self.writer.write(encoded_message)
        self.writer.drain()

    def abort(self, result_code=RESULT_PROTOCOL_ERROR, diagnostic_message=''):  # unsolicited notification of disconnection
        result = build_ldap_result(result_code, diagnostic_message=diagnostic_message)
        response = build_extended_response(result, '1.3.6.1.4.1.1466.20036')
        ldap_message = build_ldap_message(0, 'extendedResp', response)
        self.send(ldap_message)
        self.writer.close()

    def start_tls(self):
        if not self.tls_started:
            logging.debug('start_tls')
            ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
            ssl_context.load_cert_chain(self.dsa.cert_file, keyfile=self.dsa.key_file, password=self.dsa.key_file_password)
            wrapped_socket = ssl_context.wrap_socket(self.writer.get_extra_info('socket'), server_side=True, do_handshake_on_connect=False)
            wrapped_socket.do_handshake(block=True)
            return True

        return False



"""
"""

# Created on 2015.03.18
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import logging

from .. import NATIVE_ASYNCIO

if NATIVE_ASYNCIO:
    import asyncio
else:
    import trollius as asyncio
    from trollius import From, Return

from ldap3 import RESULT_INVALID_CREDENTIALS, RESULT_SUCCESS, RESULT_PROTOCOL_ERROR, RESULT_AUTH_METHOD_NOT_SUPPORTED

from ..protocol.rfc4511 import build_ldap_result, build_bind_response

# BindRequest ::= [APPLICATION 0] SEQUENCE {
#     version                 INTEGER (1 ..  127),
#     name                    LDAPDN,
#     authentication          AuthenticationChoice }
#
# AuthenticationChoice ::= CHOICE {
#     simple                  [0] OCTET STRING,
#                             -- 1 and 2 reserved
#     sasl                    [3] SaslCredentials,
# ... }
#
# SaslCredentials ::= SEQUENCE {
#     mechanism               LDAPString,
#     credentials             OCTET STRING OPTIONAL }


@asyncio.coroutine
def do_bind_operation(dua, message_id, dict_req):
    logging.debug('do BIND operation for DUA %s: %s' % (dua.identity, str(dict_req)))
    while len(dua.pending) > 1:  # wait until only the bind operation is in the pending dict
        asyncio.sleep(0.1)

    server_sasl_credentials = None
    if dict_req['version'] != 3:  # protocol version check (RFC4511 4.2 - line 878)
        result = build_ldap_result(RESULT_PROTOCOL_ERROR, diagnostic_message='only LDAP version 3 protocol allowed')
        dua.user = dua.dsa.user_backend.anonymous()
    else:
        if dict_req['authentication']['simple'] == '' and not dict_req['name']:  # anonymous simple authentication (RFC4511 4.2 - line 883)
            dua.user = dua.dsa.user_backend.anonymous()
            result = build_ldap_result(RESULT_SUCCESS, diagnostic_message='anonymous authentication successful')
        elif dict_req['name'] and dict_req['authentication']['simple']:  # simple authentication (RFC4511 4.2 - line 888)
            dua.user = dua.dsa.user_backend.find_user(dict_req['name'])
            if dua.user:
                if not dua.dsa.user_backend.check_credentials(dua.user, dict_req['authentication']['simple']):
                    if NATIVE_ASYNCIO:
                        yield from asyncio.sleep(3)  # pause if invalid user
                    else:
                        yield From(asyncio.sleep(3))  # pause if invalid user
                    result = build_ldap_result(RESULT_INVALID_CREDENTIALS, diagnostic_message='invalid credentials')
                    dua.user = dua.dsa.user_backend.anonymous()
                else:  # successful simple authentication
                    result = build_ldap_result(RESULT_SUCCESS, diagnostic_message='user authentication successful')
            else:
                if NATIVE_ASYNCIO:
                    yield from asyncio.sleep(3)  # pause if not existent user
                else:
                    yield From(asyncio.sleep(3))  # pause if not existent user
                result = build_ldap_result(RESULT_INVALID_CREDENTIALS, diagnostic_message='user not found')
                dua.user = dua.dsa.user_backend.anonymous()
        elif dict_req['authentication']['sasl']:  # sasl authentication
            result = build_ldap_result(RESULT_AUTH_METHOD_NOT_SUPPORTED, diagnostic_message='SASL not available')
            dua.user = dua.dsa.user_backend.anonymous()
        else:  # undefined
            dua.abort()
            if NATIVE_ASYNCIO:
                return None, None
            else:
                raise Return((None, None))
    response = build_bind_response(result, server_sasl_credentials)
    if NATIVE_ASYNCIO:
        return response, 'bindResponse'
    else:
        raise Return((response, 'bindResponse'))



"""
"""

# Created on
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.


class UserBaseBackend(object):
    def __init__(self):
        pass

    def find_user(self, identity):
        raise NotImplementedError

    def add_user(self, identity, authorization, credentials):
        raise NotImplementedError

    def del_user(self, identity):
        raise NotImplementedError

    def modify_user(self, identity, new_identity):
        raise NotImplementedError

    def check_credentials(self, user, credentials):
        raise NotImplementedError

    @classmethod
    def unauthenticated(cls):
        raise NotImplementedError

    @classmethod
    def anonymous(cls):
        raise NotImplementedError



"""
"""

# Created on 2015.03.15
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.
from .. import NATIVE_ASYNCIO

if NATIVE_ASYNCIO:
    import asyncio
else:
    import trollius as asyncio
    from trollius import From, Return

import ssl
import logging
from time import sleep

from pyasn1.codec.ber import decoder, encoder
from ldap3 import RESULT_SUCCESS
from ldap3.core.exceptions import LDAPExceptionError
from ldap3.strategy.base import BaseStrategy
from ldap3.operation.bind import bind_request_to_dict
from ldap3.operation.add import add_request_to_dict
from ldap3.operation.compare import compare_request_to_dict
from ldap3.operation.delete import delete_request_to_dict
from ldap3.operation.extended import extended_request_to_dict
from ldap3.operation.modify import modify_request_to_dict
from ldap3.operation.modifyDn import modify_dn_request_to_dict
from ldap3.operation.search import search_request_to_dict
from ldap3.protocol.rfc4511 import LDAPMessage
from ldap3.protocol.rfc2696 import RealSearchControlValue
from ldap3.protocol.oid import Oids

from .dua import Dua
from ..operation.bind import do_bind_operation
from ..operation.extended import do_extended_operation
from ..operation.unbind import do_unbind_operation
from ..protocol.rfc4511 import build_ldap_message


class Dsa(object):
    def __init__(self, name, address, port=389, secure_port=636, cert_file=None, key_file=None, key_file_password=None, user_backend=None):
        self.clients = dict()
        self.server = None
        self.secure_server = None
        self.instance = None
        self.name = name
        self.address = address
        self.port = port
        self.cert_file = cert_file
        self.key_file = key_file
        self.key_file_password = key_file_password
        self.user_backend = user_backend
        self.secure_port = secure_port if self.cert_file else None

    # @asyncio.coroutine
    # def status(self):
    #     last = None
    #     trigger = False
    #     while True:
    #         if len(self.clients) != last:
    #             print('Clients on DSA ', self.name + ':', len(self.clients))
    #             last = len(self.clients)
    #         if NATIVE_ASYNCIO:
    #             yield from asyncio.sleep(2)
    #         else:
    #             yield From(asyncio.sleep(2))
    #         if self.clients:
    #             trigger = True
    #         if trigger and not self.clients:
    #             break
    #     print('Closing DSA', self.name)
    #     self.stop()
    #     print('DSA {} closed'.format(self.name))

    def stop(self):
        logging.info('stopping DSA %s' % self.name)
        if self.port:
            logging.debug('closing unsecure server for DSA %s' % self.name)
            while not self.server:  # wait for server object to appear
                logging.debug('waiting for DSA %s server to appear' % self.name)
                sleep(0.2)
            self.instance.loop.call_soon_threadsafe(self.server.close)
        if self.secure_port:
            logging.debug('closing secure server for DSA %s' % self.name)
            while not self.secure_server:  # wait for secure server object to appear
                logging.debug('waiting for DSA %s secure server to appear' % self.name)
                sleep(0.2)
            self.instance.loop.call_soon_threadsafe(self.secure_server.close)

    def client_connected(self, reader, writer):
        logging.debug('client connected from')
        dua = Dua(self.user_backend.anonymous(), reader, writer, self)
        self.register_client(dua)

    def start(self):  # executed in an Executor start() method
        self.instance.loop = asyncio.new_event_loop()
        self.instance.loop.set_debug(True)
        asyncio.set_event_loop(self.instance.loop)
        logging.info('starting DSA %s' % self.name)

        if self.port:  # start unsecure server
            logging.debug('starting unsecure server for DSA %s' % self.name)
            coro = asyncio.start_server(self.client_connected, self.address, self.port)
            self.server = self.instance.loop.run_until_complete(coro)
            logging.debug('started unsecure server for DSA %s: %s' % (self.name, self.server))

        if self.secure_port:  # start secure server
            logging.debug('starting secure server for DSA %s' % self.name)
            ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
            ssl_context.load_cert_chain(self.cert_file, keyfile=self.key_file, password=self.key_file_password)
            secure_coro = asyncio.start_server(self.client_connected, self.address, self.secure_port, ssl=ssl_context)
            self.secure_server = self.instance.loop.run_until_complete(secure_coro)
            logging.debug('started secure server for DSA %s: %s' % (self.name, self.server))

        logging.info('DSA {} started'.format(self.name))

        if self.port:
            self.instance.loop.run_until_complete(self.server.wait_closed())
        if self.secure_port:
            self.instance.loop.run_until_complete(self.secure_server.wait_closed())

    @asyncio.coroutine
    def handle_client(self, dua):
        data = -1  # enter loop
        while data:
            messages = []
            receiving = True
            unprocessed = b''
            data = b''
            get_more_data = True
            while receiving:
                if get_more_data:
                    if NATIVE_ASYNCIO:
                        data = yield from dua.reader.read(4096)
                    else:
                        data = yield From(dua.reader.read(4096))

                    unprocessed += data
                if len(data) > 0:
                    length = BaseStrategy.compute_ldap_message_size(unprocessed)
                    if length == -1:  # too few data to decode message length
                        get_more_data = True
                        continue
                    if len(unprocessed) < length:
                        get_more_data = True
                    else:
                        messages.append(unprocessed[:length])
                        unprocessed = unprocessed[length:]
                        get_more_data = False
                        if len(unprocessed) == 0:
                            receiving = False
                else:
                    receiving = False
            logging.debug('received {} bytes for server {}'.format(len(data), self.name))
            if messages:
                for request in messages:
                    while len(request) > 0:
                        ldap_req, unprocessed = decoder.decode(request, asn1Spec=LDAPMessage())
                        request = unprocessed
                        self.instance.loop.create_task(self.perform_request(dua, ldap_req))
                    logging.debug('processed request for server %s' % self.name)
        logging.info('exit handle for server %s' % self.name)

    @asyncio.coroutine
    def perform_request(self, dua, request):
        message_id = int(request.getComponentByName('messageID'))
        dict_req = BaseStrategy.decode_request(request)
        if message_id not in dua.pending:
            dua.pending[message_id] = dict_req
            if dict_req['type'] == 'bindRequest':
                if NATIVE_ASYNCIO:
                    response, response_type = yield from do_bind_operation(dua, message_id, dict_req)
                else:
                    response, response_type = yield From(do_bind_operation(dua, message_id, dict_req))
            elif dict_req['type'] == 'unbindRequest':
                if NATIVE_ASYNCIO:
                    yield from do_unbind_operation(dua, message_id)
                else:
                    yield From(do_unbind_operation(dua, message_id))
                dua.writer.close()
                if NATIVE_ASYNCIO:
                    return
                else:
                    raise Return()
            elif dict_req['type'] == 'extendedReq':
                response, response_type = do_extended_operation(dua, message_id, dict_req)
                if response['responseName'] == '1.3.6.1.4.1.1466.20037' and response['result'] == RESULT_SUCCESS:  # issue start_tls
                    ldap_message = build_ldap_message(message_id, response_type, response, None)
                    dua.send(ldap_message)
                    dua.start_tls()
                response = None
            else:
                dua.abort(diagnostic_message='unknown operation')
                if NATIVE_ASYNCIO:
                    return
                else:
                    raise Return()

            del dua.pending[message_id]
            if not response:  # notice of disconnection sent while doing operation
                if NATIVE_ASYNCIO:
                    return
                else:
                    raise Return()
            controls = None  # TODO
            ldap_message = build_ldap_message(message_id, response_type, response, controls)
        else:  # pending message with same id of previous message
            dua.abort(diagnostic_message='duplicate message ID')
            if NATIVE_ASYNCIO:
                return
            else:
                raise Return()
        dua.send(ldap_message)

    def register_client(self, dua):
        task = self.instance.loop.create_task(self.handle_client(dua))
        logging.debug('new connection on DSA %s' % self.name)
        self.clients[task] = dua

        def client_done(task_done):
            logging.debug('closing connection on server %s for identity %s' % (self.name, dua.user.identity))
            self.unregister_client(task_done)
            dua.writer.close()

        task.add_done_callback(client_done)

    def unregister_client(self, task):
        del self.clients[task]



"""
"""

# Created on 2015.03.15
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.


import json
from ..user import UserBaseBackend
from ...core.user import User


class JsonUserBackend(UserBaseBackend):
    def __init__(self, json_file):
        self.json_file = json_file
        try:
            self.users = json.load(open(json_file))
        except Exception:
            self.users = {}

        super().__init__()

    def find_user(self, identity):
        if identity in self.users:
            return User(identity, self.users[identity])

        return None

    def add_user(self, identity, authorization, credentials):
        if identity not in self.users:
            self.users[identity] = {'auth': authorization, 'password': credentials}
            return True

        return False

    def del_user(self, identity):
        if identity in self.users:
            del self.users[identity]
            return True

        return False

    def modify_user(self, identity, new_identity):
        if identity in self.users:
            if self.add_user(new_identity, self.users[identity]):
                self.del_user(identity)
                return True

        return False

    def check_credentials(self, user, credentials):
        if user.identity in self.users:
            if self.users[user.identity]['password'] == credentials:
                return True

        return False

    def store(self):
        json.dump({'users': self.users}, open(self.json_file, 'w+'))

    @classmethod
    def unauthenticated(cls):
        return User('un-authenticated')

    @classmethod
    def anonymous(cls):
        return User('anonymous')



"""
"""

# Created on 2015.03.22
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import logging

from .. import NATIVE_ASYNCIO

if NATIVE_ASYNCIO:
    import asyncio
else:
    import trollius as asyncio
    from trollius import From, Return


# ExtendedRequest ::= [APPLICATION 23] SEQUENCE {
#     requestName      [0] LDAPOID,
#     requestValue     [1] OCTET STRING OPTIONAL }

from ldap3 import RESULT_SUCCESS, RESULT_PROTOCOL_ERROR, RESULT_UNAVAILABLE

from ..protocol.rfc4511 import build_ldap_result, build_extended_response

@asyncio.coroutine
def do_extended_operation(dua, message_id, dict_req):
    logging.debug('do EXTENDED operation for DUA %s: %s' % (dua.identity, str(dict_req)))

    if dict_req['name'] == '1.3.6.1.4.1.1466.20037':  # start_tls
        if dua.dsa.secure_port:
            result = build_ldap_result(RESULT_SUCCESS)
        else:
            result = build_ldap_result(RESULT_UNAVAILABLE)
        response = build_extended_response(result, '1.3.6.1.4.1.1466.20037')
    else:
        result = build_ldap_result(RESULT_PROTOCOL_ERROR, diagnostic_message='extended operation not supported')
        response = build_extended_response(result)

    if NATIVE_ASYNCIO:
        return response, 'extendedResp'
    else:
        raise Response((response, 'extendedRespo'))



NATIVE_ASYNCIO = False

try:
    # Use builtin asyncio
    from asyncio import BaseEventLoop
    NATIVE_ASYNCIO = True
except ImportError:
    # Use Trollius for backward compatability
    import trollius
    from trollius import From, Return


EXEC_PROCESS = 'PROCESS'
EXEC_THREAD = 'THREAD'

from .version import __author__, __version__, __email__, __description__, __status__, __license__, __url__

from .core.dsa import Dsa
from .core.instance import Instance
from .backend.user.json import JsonUserBackend




"""
"""

# Created on 2015.03.22
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of sldap3.
#
# sldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# sldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with sldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

from ldap3.protocol.rfc4511 import LDAPResult, LDAPMessage, ProtocolOp, MessageID, Referral, BindResponse, \
    ServerSaslCreds, ExtendedResponse
from ldap3.protocol.convert import build_controls_list


def build_ldap_message(message_id, response_type, response, controls=None):
    # LDAPMessage ::= SEQUENCE {
    # messageID       MessageID,
    #     protocolOp      CHOICE {
    #         bindRequest           BindRequest,
    #         bindResponse          BindResponse,
    #         unbindRequest         UnbindRequest,
    #         searchRequest         SearchRequest,
    #         searchResEntry        SearchResultEntry,
    #         searchResDone         SearchResultDone,
    #         searchResRef          SearchResultReference,
    #         modifyRequest         ModifyRequest,
    #         modifyResponse        ModifyResponse,
    #         addRequest            AddRequest,
    #         addResponse           AddResponse,
    #         delRequest            DelRequest,
    #         delResponse           DelResponse,
    #         modDNRequest          ModifyDNRequest,
    #         modDNResponse         ModifyDNResponse,
    #         compareRequest        CompareRequest,
    #         compareResponse       CompareResponse,
    #         abandonRequest        AbandonRequest,
    #         extendedReq           ExtendedRequest,
    #         extendedResp          ExtendedResponse,
    #         ...,
    #         intermediateResponse  IntermediateResponse },
    #     controls       [0] Controls OPTIONAL }

    ldap_message = LDAPMessage()
    ldap_message['messageID'] = MessageID(message_id)
    ldap_message['protocolOp'] = ProtocolOp().setComponentByName(response_type, response)
    message_controls = build_controls_list(controls)
    if message_controls is not None:
        ldap_message['controls'] = message_controls

    return ldap_message


def build_ldap_result(result_code, matched_dn='', diagnostic_message='', referral=None):
    # LDAPResult ::= SEQUENCE {
    #     resultCode         ENUMERATED {
    #         success                      (0),
    #         operationsError              (1),
    #         protocolError                (2),
    #         timeLimitExceeded            (3),
    #         sizeLimitExceeded            (4),
    #         compareFalse                 (5),
    #         compareTrue                  (6),
    #         authMethodNotSupported       (7),
    #         strongerAuthRequired         (8),
    #              -- 9 reserved --
    #         referral                     (10),
    #         adminLimitExceeded           (11),
    #         unavailableCriticalExtension (12),
    #         confidentialityRequired      (13),
    #         saslBindInProgress           (14),
    #         noSuchAttribute              (16),
    #         undefinedAttributeType       (17),
    #         inappropriateMatching        (18),
    #         constraintViolation          (19),
    #         attributeOrValueExists       (20),
    #         invalidAttributeSyntax       (21),
    #              -- 22-31 unused --
    #         noSuchObject                 (32),
    #         aliasProblem                 (33),
    #         invalidDNSyntax              (34),
    #              -- 35 reserved for undefined isLeaf --
    #         aliasDereferencingProblem    (36),
    #              -- 37-47 unused --
    #         inappropriateAuthentication  (48),
    #         invalidCredentials           (49),
    #         insufficientAccessRights     (50),
    #         busy                         (51),
    #         unavailable                  (52),
    #         unwillingToPerform           (53),
    #         loopDetect                   (54),
    #              -- 55-63 unused --
    #         namingViolation              (64),
    #         objectClassViolation         (65),
    #         notAllowedOnNonLeaf          (66),
    #         notAllowedOnRDN              (67),
    #         entryAlreadyExists           (68),
    #         objectClassModsProhibited    (69),
    #              -- 70 reserved for CLDAP --
    #         affectsMultipleDSAs          (71),
    #              -- 72-79 unused --
    #         other                        (80),
    #         ...  },
    #     matchedDN          LDAPDN,
    #     diagnosticMessage  LDAPString,
    #     referral           [3] Referral OPTIONAL }

    ldap_result = LDAPResult()
    ldap_result['resultCode'] = result_code
    ldap_result['matchedDN'] = matched_dn
    ldap_result['diagnosticMessage'] = diagnostic_message
    if referral:
        ldap_result['referral'] = Referral(referral)
    return ldap_result


def build_bind_response(ldap_result, server_sasl_credentials):
    # BindResponse ::= [APPLICATION 1] SEQUENCE {
    #     COMPONENTS OF LDAPResult,
    #     serverSaslCreds    [7] OCTET STRING OPTIONAL }

    response = BindResponse()
    response['resultCode'] = ldap_result['resultCode']
    response['matchedDN'] = ldap_result['matchedDN']
    response['diagnosticMessage'] = ldap_result['diagnosticMessage']
    if ldap_result['referral']:
        response['referral'] = ldap_result['referral']
    if server_sasl_credentials:
        response['serverSaslCreds'] = ServerSaslCreds(server_sasl_credentials)

    return response


def build_extended_response(ldap_result, response_name=None, response_value=None):
    # ExtendedResponse ::= [APPLICATION 24] SEQUENCE {
    #     COMPONENTS OF LDAPResult,
    #     responseName     [10] LDAPOID OPTIONAL,
    #     responseValue    [11] OCTET STRING OPTIONAL }
    response = ExtendedResponse()
    response['resultCode'] = ldap_result['resultCode']
    response['matchedDN'] = ldap_result['matchedDN']
    response['diagnosticMessage'] = ldap_result['diagnosticMessage']
    if response_name:
        response['responseName'] = response_name
    if response_value:
        response['responseValue'] = response_value

    return response



"""
"""

# Created on 2015.02.06
#
# Author: Giovanni Cannata
#
# Copyright 2015 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import logging
from os import linesep
from threading import Lock
from datetime import datetime
from pyasn1.codec.ber import encoder

from ldap3.core.exceptions import LDAPExceptionError, LDAPUnknownRequestError
from ldap3 import RESULT_OPERATIONS_ERROR, RESULT_SUCCESS, RESULT_AUTH_METHOD_NOT_SUPPORTED, RESULT_INVALID_CREDENTIALS, \
    SIMPLE, ANONYMOUS, RESTARTABLE, RESULT_PROTOCOL_ERROR
from ldap3.operation.bind import bind_response_operation
from ldap3.strategy.base import BaseStrategy
from ldap3.utils.ciDict import CaseInsensitiveDict
from ldap3.protocol.rfc4511 import LDAPMessage, MessageID


class TreeNode(object):
    def __init__(self, tag=None, assertion=None):
        self.tag = tag
        self.parent = None
        self.assertion = assertion
        self.elements = []

    def append(self, filter_node):
        filter_node.parent = self
        self.elements.append(filter_node)
        return filter_node

    def __str__(self, pos=0):
        self.__repr__(pos)

    def __repr__(self, pos=0):
        node_tags = ['ROOT', 'AND', 'OR', 'NOT', 'MATCH_APPROX', 'MATCH_GREATER_OR_EQUAL', 'MATCH_LESS_OR_EQUAL', 'MATCH_EXTENSIBLE', 'MATCH_PRESENT', 'MATCH_SUBSTRING', 'MATCH_EQUAL']
        representation = ' ' * pos + 'tag: ' + node_tags[self.tag] + ' - assertion: ' + str(self.assertion)
        if self.elements:
            representation += ' - elements: ' + str(len(self.elements))
            for element in self.elements:
                representation += linesep + ' ' * pos + element.__repr__(pos + 2)
        return representation


class Client(object):
    def __init__(self, connection, authentication_type, user=None):
        self.bind_time = datetime.now()
        self.auth_type = authentication_type
        self.user = user
        self.connection = connection


class Dsa(object):
    def __init__(self, server):
        self.database = CaseInsensitiveDict()
        self.connections = dict()
        self.ready_to_send = dict()
        self.lock = Lock()
        self.server = server

    def produce_response(self, connection_id):
        message_id = min(self.ready_to_send.keys())
        logging.debug('SERVER: processing response ' + message_id)
        if message_id in self.ready_to_send:
            ldap_message = LDAPMessage()
            ldap_message['messageID'] = MessageID(message_id)
            ldap_message['protocolOp'] = self.ready_to_send.pop(message_id)
            encoded_message = encoder.encode(ldap_message)
            with self.lock:
                return encoded_message
        else:
            raise LDAPExceptionError('response not ready in mock server')

    def accept_request(self, connection_id, ldap_message):
        request = BaseStrategy.decode_request(ldap_message)
        logging.debug('SERVER: processing request ' + request + ' for ' + connection_id)

        response = None
        if request['type'] == 'bindRequest':
            response = self.do_bind(connection_id, request)
        elif request['type'] == 'unbindRequest':
            self.do_unbind(connection_id, request)
            pass  # unbind doesn't return anything
        elif request['type'] == 'addRequest':
            response = self.do_add(connection_id, request)
        elif request['type'] == 'compareRequest':
            response = self.do_compare(connection_id, request)
        elif request['type'] == 'delRequest':
            response = self.do_delete(connection_id, request)
        elif request['type'] == 'extendedReq':
            response = self.do_extended(connection_id, request)
        elif request['type'] == 'modifyRequest':
            response = self.do_modify(connection_id, request)
        elif request['type'] == 'modDNRequest':
            response = self.do_modify_dn(connection_id, request)
        elif request['type'] == 'searchRequest':
            response = self.do_search(connection_id, request)
        elif request['type'] == 'abandonRequest':
            response = self.do_abandon(connection_id, request)
        else:
            raise LDAPUnknownRequestError('unknown request')

        if response:
            self.ready_to_send[int(ldap_message['messageID'])] = response

    def do_open(self, connection_id):
        client = Client(connection=connection_id, authentication_type=ANONYMOUS)
        with self.lock:
            self.connections[connection_id] = client

    def do_bind(self, connection_id, request):
        if request['version'] == 3:
            if 'simple' in request['authentication']:
                if request['authentication']['simple'] == 'mock_password':
                    result_code = RESULT_SUCCESS
                    diagnostic_message='bind successful'
                    if connection_id in self.connections:
                        with self.lock:
                            self.connections[connection_id].user = request['name']
                    else:
                        result_code = RESULT_PROTOCOL_ERROR
                        diagnostic_message = 'connection not open'
                else:
                    result_code = RESULT_INVALID_CREDENTIALS
                    diagnostic_message = 'bind unsuccessful'
            else:
                result_code = RESULT_AUTH_METHOD_NOT_SUPPORTED
                diagnostic_message = 'only simple authentication allowed'
        else:
            result_code = RESULT_OPERATIONS_ERROR
            diagnostic_message = 'only version 3 of LDAP protocol is supported'

        response = bind_response_operation(result_code=result_code, diagnostic_message=diagnostic_message)
        return response

    def do_unbind(self, connection_id, request):
        with self.lock:
            del self.connections[connection_id]

    def do_add(self, connection_id, request):
        with self.lock:
            result = None
            return result

    def do_compare(self, connection_id, request):
        with self.lock:
            result = None
            return result

    def do_delete(self, connection_id, request):
        with self.lock:
            result = None
            return result

    def do_modify(self, connection_id, request):
        with self.lock:
            result = None
            return result

    def do_modify_dn(self, connection_id, request):
        with self.lock:
            result = None
            return result

    def do_search(self, connection_id, request):
        with self.lock:
            response = None
            return response

    def do_abandon(self, connection_id, request):
        with self.lock:
            result = None
            return result

    def do_extended(self, connection_id, request):
        with self.lock:
            result = None
            return result

